/*
** Job Arranger for ZABBIX
** Copyright (C) 2012 FitechForce, Inc. All Rights Reserved.
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
**/

/*
** $Date:: 2012-11-29 18:25:26 +0900 #$
** $Revision: 2874 $
** $Author: ossinfra@FITECHLABS.CO.JP $
**/

#include <json/json.h>
#include "common.h"
#include "log.h"

#if defined(ZABBIX_SERVICE)
#include "service.h"
#elif defined(ZABBIX_DAEMON)
#include "daemon.h"
#endif

#include "jacommon.h"
#include "jafile.h"
#include "jajobfile.h"
#include "jajobobject.h"
#include "jakill.h"
#include "jastr.h"
#include "jadb.h"
#include "jajobdb.h"
#include "jaagent.h"
#include "jatelegram.h"
#include "jatcp.h"

char *jobext[] = { "start", "end", "ret", "stdout", "stderr", NULL };

const int check_cnt = 30;

/******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *                                                                            *
 * Purpose:                                                                   *
 *                                                                            *
 * Parameters:                                                                *
 *                                                                            *
 * Return value:                                                              *
 *                                                                            *
 * Comments:                                                                  *
 *                                                                            *
 ******************************************************************************/
int ja_agent_setenv(ja_job_object * job, char *env_string)
{
    int ret;
    json_object *json_env;
    char *key;
    char tmp_string[JA_MAX_DATA_LEN];
    struct json_object *val;
    struct lh_entry *entry;
    const char *__function_name = "ja_job_object_setenv";

    if (job == NULL)
        return FAIL;
    zabbix_log(LOG_LEVEL_DEBUG, "In %s() jobid: " ZBX_FS_UI64 ", env: %s",
               __function_name, job->jobid, job->env);

    memset(tmp_string, 0, JA_MAX_DATA_LEN);
    if (env_string != NULL)
        memset(env_string, 0, JA_MAX_DATA_LEN);

    if (strlen(job->env) == 0)
        return SUCCEED;

    ret = FAIL;
    json_env = json_tokener_parse(job->env);
    if (is_error(json_env)) {
        zbx_snprintf(job->message, sizeof(job->message),
                     "Can not parse json env data [%s]", job->env);
        json_env = NULL;
        goto clear;
    }

    for (entry = json_object_get_object(json_env)->head;
         (entry
          ? (key = (char *) entry->k, val =
             (struct json_object *) entry->v, entry) : 0);
         entry = entry->next) {
        zabbix_log(LOG_LEVEL_DEBUG, "setenv %s=%s", key,
                   json_object_get_string(val));
#ifdef _WINDOWS
        zbx_snprintf(tmp_string, JA_MAX_DATA_LEN, "%s", env_string);
        zbx_snprintf(env_string, JA_MAX_DATA_LEN, "%sset %s=%s\r\n",
                     tmp_string, key, json_object_get_string(val));
#else
        if (setenv(key, json_object_get_string(val), 1) != 0) {
            zbx_snprintf(job->message, sizeof(job->message),
                         "Can not setenv %s=%s", key,
                         json_object_get_string(val));
            goto clear;
        }
#endif
    }

    ret = SUCCEED;
  clear:
    if (json_env != NULL)
        json_object_put(json_env);
    if (ret == FAIL)
        zabbix_log(LOG_LEVEL_ERR, "%s", job->message);
    return ret;
}

/******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *                                                                            *
 * Purpose:                                                                   *
 *                                                                            *
 * Parameters:                                                                *
 *                                                                            *
 * Return value:                                                              *
 *                                                                            *
 * Comments:                                                                  *
 *                                                                            *
 ******************************************************************************/
int ja_agent_kill(ja_job_object * job)
{
    int ret, status;
    JA_PID pid;
    const char *__function_name = "ja_agent_kill";

    if (job == NULL)
        return FAIL;
    zabbix_log(LOG_LEVEL_DEBUG, "In %s() jobid: " ZBX_FS_UI64,
               __function_name, job->jobid);

    ret = FAIL;
    if (ja_db_begin() != SUCCEED) {
        zbx_snprintf(job->message, sizeof(job->message),
                     "Can not kill jobid:", ZBX_FS_UI64, job->jobid);
        goto close;
    }

    status = ja_jobdb_get_status(job->jobid);
    if (status == JA_AGENT_STATUS_RUN) {
        pid = ja_jobdb_get_pid(job->jobid);
        if (ja_kill(pid) == FAIL) {
            zbx_snprintf(job->message, sizeof(job->message),
                         "Can not kill pid: %d, jobid: " ZBX_FS_UI64,
                         pid, job->jobid);
            goto close;
        }
    } else {
        zbx_snprintf(job->message, sizeof(job->message),
                     "jobid: " ZBX_FS_UI64 " is not running", job->jobid);
        goto close;
    }

    ret = SUCCEED;
  close:
    ja_db_commit();
    if (ret == FAIL) {
        job->result = JA_JOBRUN_RES_FAIL;
        zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name,
                   job->message);
    }
    return ret;
}

/******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *                                                                            *
 * Purpose:                                                                   *
 *                                                                            *
 * Parameters:                                                                *
 *                                                                            *
 * Return value:                                                              *
 *                                                                            *
 * Comments:                                                                  *
 *                                                                            *
 ******************************************************************************/
int ja_agent_begin(ja_job_object * job)
{
    const char *__function_name = "ja_agent_begin";

    if (job == NULL)
        return FAIL;
    zabbix_log(LOG_LEVEL_INFORMATION,
               "In %s() jobid: " ZBX_FS_UI64 ", method: %d",
               __function_name, job->jobid, job->method);

    switch (job->method) {
    case JA_AGENT_METHOD_NORMAL:
    case JA_AGENT_METHOD_ABORT:
    case JA_AGENT_METHOD_TEST:
        job->status = JA_AGENT_STATUS_BEGIN;
        if (ja_jobdb_insert(job) == FAIL)
            goto error;
        if (ja_file_create(CONFIG_REQUEST_FLAG, 1) == FAIL) {
            job->result = JA_JOBRUN_RES_FAIL;
            zbx_snprintf(job->message, sizeof(job->message),
                         "Can not create the file [%s]",
                         CONFIG_REQUEST_FLAG);
            goto error;
        }
        break;
    case JA_AGENT_METHOD_KILL:
        if (ja_agent_kill(job) == FAIL) {
            goto error;
        }
        break;
    default:
        job->result = JA_JOBRUN_RES_FAIL;
        zbx_snprintf(job->message, sizeof(job->message),
                     "Invalid method: %d, jobid " ZBX_FS_UI64 ")",
                     job->method, job->jobid);
        goto error;
        break;
    }

    zabbix_log(LOG_LEVEL_INFORMATION,
               "jobid: " ZBX_FS_UI64 ", method: %d is begin", job->jobid,
               job->method);
    return SUCCEED;

  error:
    zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name, job->message);
    return FAIL;
}

/******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *                                                                            *
 * Purpose:                                                                   *
 *                                                                            *
 * Parameters:                                                                *
 *                                                                            *
 * Return value:                                                              *
 *                                                                            *
 * Comments:                                                                  *
 *                                                                            *
 ******************************************************************************/
int ja_agent_run(ja_job_object * job)
{
    char filepath[JA_MAX_STRING_LEN];
    char full_command[JA_MAX_STRING_LEN];
    const char *__function_name = "ja_agent_run";
#ifdef _WINDOWS
    int ret;
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    LPTSTR wcommand;
    char env[JA_MAX_DATA_LEN], full_script[JA_MAX_DATA_LEN];

    zabbix_log(LOG_LEVEL_DEBUG, "In %s()", __function_name);
    if (job == NULL)
        return FAIL;
    zabbix_log(LOG_LEVEL_DEBUG,
               "jobid: " ZBX_FS_UI64 ", status: %d", job->jobid,
               job->status);

    ret = FAIL;
    memset(&si, 0, sizeof(si));
    si.cb = sizeof(si);
    GetStartupInfo(&si);
    wcommand = NULL;
    job->start_time = time(NULL);
    job->end_time = time(NULL);

    zbx_snprintf(filepath, JA_MAX_STRING_LEN, "%s%c%s-" ZBX_FS_UI64,
                 CONFIG_TMPDIR, JA_DLM, progname, job->jobid);
    if (ja_agent_setenv(job, env) == FAIL)
        goto exit;
    if (job->method == JA_AGENT_METHOD_TEST)
        zbx_snprintf(full_script, JA_MAX_DATA_LEN, "%s", env);
    else
        zbx_snprintf(full_script, JA_MAX_DATA_LEN, "%s%s", env,
                     job->script);
    if (ja_jobfile_create(filepath, jobext, full_script) == FAIL) {
        zbx_snprintf(job->message, sizeof(job->message),
                     "Can not create the result files [%s]", filepath);
        goto exit;
    }
    zbx_snprintf(full_command, JA_MAX_STRING_LEN,
                 "\"%s\" \"%s\" \"%s.%s\"", CONFIG_CMD_FILE, filepath,
                 filepath, JA_EXE);
    wcommand = zbx_acp_to_unicode(full_command);
    if (0 == CreateProcess(NULL,        /* no module name (use command line) */
                           wcommand,    /* name of app to launch */
                           NULL,        /* default process security attributes */
                           NULL,        /* default thread security attributes */
                           FALSE,       /* do not inherit handles from the parent */
                           0,   /* normal priority */
                           NULL,        /* use the same environment as the parent */
                           NULL,        /* launch in the current directory */
                           &si, /* startup information */
                           &pi  /* process information stored upon return */
        )) {
        zbx_snprintf(job->message, sizeof(job->message),
                     "failed to create process for jobid " ZBX_FS_UI64
                     " [%s]: %s", job->jobid, full_command,
                     strerror_from_system(GetLastError()));
        goto exit;
    } else {
        job->pid = pi.dwProcessId;
        CloseHandle(pi.hProcess);
        CloseHandle(pi.hThread);
        ret = SUCCEED;
    }

  exit:
    if (ret == FAIL) {
        job->result = JA_JOBRESULT_FAIL;
        job->status = JA_AGENT_STATUS_END;
        zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name,
                   job->message);
    } else {
        job->status = JA_AGENT_STATUS_RUN;
    }
    ja_jobdb_update(job);
    if (wcommand != NULL)
        zbx_free(wcommand);
    return ret;
#else
    pid_t pid;
    int ret, i;
    zabbix_log(LOG_LEVEL_DEBUG, "In %s()", __function_name);

    zabbix_log(LOG_LEVEL_INFORMATION,
               "In %s() jobid: " ZBX_FS_UI64 ", status: %d",
               __function_name, job->jobid, job->status);

    job->start_time = time(NULL);
    job->end_time = time(NULL);

    pid = zbx_fork();
    if (pid == -1) {
        job->result = JA_JOBRESULT_FAIL;
        job->status = JA_AGENT_STATUS_END;
        zbx_snprintf(job->message, sizeof(job->message),
                     "first fork() failed for jobid: " ZBX_FS_UI64
                     "[%s]", job->jobid, zbx_strerror(errno));
        ja_jobdb_update(job);
        zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name,
                   job->message);
        return FAIL;
    } else if (pid != 0) {
        waitpid(pid, NULL, 0);
        for (i = 0; i < check_cnt; i++) {
            zbx_sleep(1);
            ja_jobdb_load_jobid(job->jobid, job);
            if (job->status != JA_AGENT_STATUS_BEGIN) {
                return SUCCEED;
            }
        }
        job->status = JA_AGENT_STATUS_END;
        job->result = JA_JOBRESULT_FAIL;
        zbx_snprintf(job->message, sizeof(job->message),
                     "can not start jobid: " ZBX_FS_UI64, job->jobid);
        ja_jobdb_update(job);
        zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name,
                   job->message);

        return FAIL;
    }

    pid = zbx_fork();
    if (pid == -1) {
        job->result = JA_JOBRESULT_FAIL;
        job->status = JA_AGENT_STATUS_END;
        zbx_snprintf(job->message, sizeof(job->message),
                     "second fork() failed for jobid: " ZBX_FS_UI64
                     "(%s)", job->jobid, zbx_strerror(errno));
        ja_jobdb_update(job);
        exit(0);
    } else if (pid != 0) {
        waitpid(pid, NULL, WNOHANG);
        exit(0);
    }

    job->pid = getpid();
    job->status = JA_AGENT_STATUS_RUN;
    zbx_snprintf(filepath, JA_MAX_STRING_LEN,
                 "%s%c%s-" ZBX_FS_UI64 "-%d", CONFIG_TMPDIR, JA_DLM,
                 progname, job->jobid, job->pid);
    zbx_snprintf(full_command, JA_MAX_STRING_LEN, "%s.%s", filepath,
                 JA_EXE);

    ret = ja_agent_setenv(job, NULL);
    if (ret == SUCCEED) {
        if (job->method == JA_AGENT_METHOD_TEST) {
            ret = ja_jobfile_create(filepath, jobext, "");
        } else {
            ret = ja_jobfile_create(filepath, jobext, job->script);
        }
        if (ret == FAIL)
            zbx_snprintf(job->message, sizeof(job->message),
                         "Can not create the result files [%s]", filepath);
    }

    if (ret == FAIL) {
        job->result = JA_JOBRESULT_FAIL;
        job->status = JA_AGENT_STATUS_END;
        zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name,
                   job->message);
    }
    ja_jobdb_update(job);
    if (ret == FAIL)
        exit(0);

    execl(CONFIG_CMD_FILE, CONFIG_CMD_FILE, filepath, full_command, NULL);
    job->result = JA_JOBRESULT_FAIL;
    job->status = JA_AGENT_STATUS_END;
    zbx_snprintf(job->message, sizeof(job->message),
                 "execl() failed for [%s]: %s", CONFIG_CMD_FILE,
                 zbx_strerror(errno));
    ja_jobdb_update(job);
    zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name, job->message);
    ja_jobfile_remove(filepath, jobext);
    exit(0);
#endif
}

/******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *                                                                            *
 * Purpose:                                                                   *
 *                                                                            *
 * Parameters:                                                                *
 *                                                                            *
 * Return value:                                                              *
 *                                                                            *
 * Comments:                                                                  *
 *                                                                            *
 ******************************************************************************/
int ja_agent_end(ja_job_object * job)
{
    int ret, chk;
    char filepath[JA_MAX_STRING_LEN];

    const char *__function_name = "ja_agent_end";
    zabbix_log(LOG_LEVEL_DEBUG,
               "In %s() jobid: " ZBX_FS_UI64 ", status: %d",
               __function_name, job->jobid, job->status);
#ifdef _WINDOWS
    zbx_snprintf(filepath, JA_MAX_STRING_LEN,
                 "%s%c%s-" ZBX_FS_UI64, CONFIG_TMPDIR, JA_DLM,
                 progname, job->jobid);
#else
    zbx_snprintf(filepath, JA_MAX_STRING_LEN,
                 "%s%c%s-" ZBX_FS_UI64 "-%d", CONFIG_TMPDIR, JA_DLM,
                 progname, job->jobid, job->pid);
#endif

    chk = ja_jobfile_chkend(filepath, job->pid);
    if (chk == 0)
        return FAIL;

    ret = FAIL;
    job->status = JA_AGENT_STATUS_END;
    if (chk == -1) {
        job->result = JA_JOBRESULT_FAIL;
        zbx_snprintf(job->message, sizeof(job->message),
                     "Check job status(end) failed. jobid: " ZBX_FS_UI64,
                     job->jobid);
    } else {
        if (ja_jobfile_load(filepath, job) == FAIL) {

            job->result = JA_JOBRESULT_FAIL;
            zbx_snprintf(job->message, sizeof(job->message),
                         "Can not load the result files[%s]. jobid: "
                         ZBX_FS_UI64, filepath, job->jobid);

        } else {
            ret = SUCCEED;
        }
    }

    if (ret == FAIL)
        zabbix_log(LOG_LEVEL_ERR, "In %s() %s", __function_name,
                   job->message);

    if (ja_jobdb_update(job) != SUCCEED)
        job->status = JA_AGENT_STATUS_RUN;

    return ret;
}

/******************************************************************************
 *                                                                            *
 * Function:                                                                  *
 *                                                                            *
 * Purpose:                                                                   *
 *                                                                            *
 * Parameters:                                                                *
 *                                                                            *
 * Return value:                                                              *
 *                                                                            *
 * Comments:                                                                  *
 *                                                                            *
 ******************************************************************************/
int ja_agent_close(ja_job_object * job)
{
    char filepath[JA_MAX_STRING_LEN];
    char telegram[JA_MAX_TELEGRAM_LEN];
    const char *__function_name = "ja_agent_close";

    zabbix_log(LOG_LEVEL_INFORMATION,
               "In %s() jobid: " ZBX_FS_UI64 ", status: %d",
               __function_name, job->jobid, job->status);

    if (ja_telegram_to_jobresult(job, telegram) == FAIL)
        return FAIL;

    if (ja_tcp_send(telegram) == FAIL) {
        job->send_retry++;
        if (job->send_retry >= CONFIG_SEND_RETRY) {
            zbx_snprintf(job->message, sizeof(job->message),
                         "Can not send the result to server. jobid: "
                         ZBX_FS_UI64, job->jobid);
            job->status = JA_AGENT_STATUS_CLOSE;
            ja_jobdb_update(job);
        }
        return FAIL;
    }

    job->status = JA_AGENT_STATUS_CLOSE;
    ja_jobdb_update(job);

#ifdef _WINDOWS
    zbx_snprintf(filepath, JA_MAX_STRING_LEN,
                 "%s%c%s-" ZBX_FS_UI64, CONFIG_TMPDIR, JA_DLM,
                 progname, job->jobid);
#else
    zbx_snprintf(filepath, JA_MAX_STRING_LEN,
                 "%s%c%s-" ZBX_FS_UI64 "-%d", CONFIG_TMPDIR, JA_DLM,
                 progname, job->jobid, job->pid);
#endif

    if (job->result == JA_JOBRESULT_SUCCEED)
        ja_jobfile_remove(filepath, jobext);

    return SUCCEED;
}
